; SilentTimer Action Figure software, v1.5
; asm.S
; things that can't be done in C
; 
; 07.03.2016
; Copyright (C) 2014-2016  Balthasar Szczepański
; 
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "msp430.h"
;Why do I even have to define this?!
#define SR R2

.section .text


;function to add 2-digit BCD numbers
.global bcd_add_byte
.type bcd_add_byte, @function
bcd_add_byte:
	clrc
	dadd.b R14, R15
	ret

;function to enter low power mode 3
.global lpm3
.type lpm3, @function
lpm3:
	bis.w #SCG1|SCG0|CPUOFF, SR
	ret

;helper function - shifts R13:R12 two digits left
.global calc_shl
.type calc_shl, @function
.type cshl_1, @function
calc_shl:
	mov   #8,      R15
cshl_1:
	rla            R12
	rlc            R13
	dec            R15
	jnz            cshl_1
	ret

;helper function - shifts R13:R12 two digits right, with rounding
.global calc_shr
.type calc_shr, @function
.type cshr_1, @function
.type cshr_2, @function
calc_shr:
	mov   R12,     R15
	and   #0x00f0, R15
	cmp   #0x0050, R15
	jlo            cshr_1
	clrc
	dadd  #0x0100, R12
	dadc           R13
cshr_1:
	mov   #8,      R15
cshr_2:
	rra            R13
	rrc            R12
	dec            R15
	jnz            cshr_2
	ret

;helper function - negate bcd number in R13:R12
.global calc_neg
.type calc_neg, @function
calc_neg:
	mov   #0x9999, R15
	sub   R12,     R15
	mov   R15,     R12
	mov   #0x9999, R15
	sub   R13,     R15
	mov   R15,     R13
	clrc
	dadd  #0x0001, R12
	dadc           R13
	ret
	
;the calculator add function
.global calc_add
.type calc_add, @function
.type ca_shl2, @function
.type ca_neg2, @function
.type ca_shl1, @function
.type ca_neg1, @function
.type ca_negr, @function
.type ca_shr1r, @function
.type ca_err, @function
.type ca_end, @function
calc_add:
	push  R11
	push  R10
	
	mov   R15,     R14
	
	mov   2(R14),  R12
	clr            R13
	bit.b #0x08,   6(R14)
	jnz            ca_shl2
	call           #calc_shl
ca_shl2:
	bit.b #0x04,   6(R14)
	jz             ca_neg2
	call           #calc_neg
ca_neg2:
	mov   R12,     R10
	mov   R13,     R11
	
	mov   4(R14),  R12
	clr            R13
	bit.b #0x20,   6(R14)
	jnz            ca_shl1
	call           #calc_shl
ca_shl1:
	bit.b #0x10,   6(R14)
	jz             ca_neg1
	call           #calc_neg
ca_neg1:
;ok, prepared, now add
	clrc
	dadd  R10,     R12
	dadd  R11,     R13
;and process result
	mov   #0xff00, R10
	mov   #0x0000, R11
	bic.b #0x03,   6(R14)
	mov   R13,     R15
	and   #0xff00, R15
	cmp   #0x9900, R15
	jne            ca_negr
	mov   #0xfff0, R10
	mov   #0xf000, R11
	bis.b #0x01,   6(R14)
	call           #calc_neg
ca_negr:
	bit   R10,     R13
	jnz            ca_err
	bit   #0xffff, R13
	jnz            ca_shr1r         
	bit   R11,     R12
	jnz            ca_shr1r
	bit   #0x00ff, R12
	jz             ca_shr1r
	bis.b #0x02,   6(R14)
	jmp            ca_end
ca_shr1r:
	call           #calc_shr
	bit   #0xffff, R13
	jnz            ca_err
	bit   R11,     R12
	jnz            ca_err
	jmp            ca_end
ca_err:
	mov.b #0xff,   7(R14)
ca_end:
	mov   R12,     0(R14)
	pop            R10
	pop            R11
	ret

;the calculator multiply function
.global calc_mul
.type calc_mul, @function
.type cm_ml, @function
.type cm_slo, @function
.type cm_sla, @function
.type cm_adm1, @function
.type cm_adm2, @function
.type cm_nr1, @function
.type cm_nr2, @function
.type cm_srl1, @function
.type cm_srl2, @function
.type cm_srl3, @function
.type cm_srl4, @function
.type cm_srrl1, @function
.type cm_srrl2, @function
.type cm_tbg, @function
.type cm_srrb, @function
.type cm_zap, @function
.type cm_err, @function
.type cm_end, @function
calc_mul:
	push  R11
	push  R10
	push  R9
	push  R8
	push  R7
	push  R6
	mov   2(R15),  R7
	mov   4(R15),  R8
	clr            R9
	clr            R12
	clr            R13
	mov   #4,      R14
cm_ml:
	clr            R10
	clr            R11
	mov   #4,      R6
cm_slo:
	rla            R8
	rlc            R9
	dec            R6
	jnz            cm_slo
	mov   #4,      R6
cm_sla:
	rla            R12
	rlc            R13
	dec            R6
	jnz            cm_sla
	bit   #0xffff, R9
	jz             cm_adm2
cm_adm1:
	clrc
	dadd  R7,      R10
	dadc           R11
	dec            R9
	jnz            cm_adm1
cm_adm2:
	clrc
	dadd  R10,     R12
	dadd  R11,     R13
	dec            R14
	jnz            cm_ml
;R8, R9, R14 = 0 here
	mov   #0xff00, R7
	bic.b #0x03,   6(R15)
	bit.b #0x10,   6(R15)
	jz             cm_nr1
	xor.b #0x01,   6(R15)
	xor   #0xf000, R8
	xor   #0x00f0, R7
cm_nr1:
	bit.b #0x04,   6(R15)
	jz             cm_nr2
	xor.b #0x01,   6(R15)
	xor   #0xf000, R8
	xor   #0x00f0, R7
cm_nr2:
	bit.b #0x20,   6(R15)
	jnz            cm_srl1
	add   #8,      R9
cm_srl1:
	bit.b #0x08,   6(R15)
	jnz            cm_srl2
	add   #8,      R9
cm_srl2:
	cmp   #0x0000, R9
	jeq            cm_srl4
cm_srl3:
	rla            R12
	rlc            R13
	rlc            R14
	dec            R9
	jnz            cm_srl3
cm_srl4:
	bit   #0xffff, R14
	jnz            cm_err
	bit   R8,      R13
	jnz            cm_err
	bit   R7,      R13
	jnz            cm_srrl2
	mov   #8,      R6
	bis.b #0x02,   6(R15)
cm_srrl1:
	rla            R12
	rlc            R13
	dec            R6
	jnz            cm_srrl1
cm_srrl2:
	and   #0xf000, R12
	cmp   #0x5000, R12
	jlo            cm_zap
	clrc
	dadd  #0x0001, R13
	dadc           R14
	jnz            cm_tbg
	bit   R8,      R13
	jnz            cm_tbg
	jmp            cm_zap
cm_tbg:
	bit.b #0x02,   6(R15)
	jz             cm_err
	bic.b #0x02,   6(R15)
	mov   #8,      R6
cm_srrb:
	rra            R13
	dec            R6
	jnz            cm_srrb
cm_zap:
	bit.b #0x02,   6(R15)
	jz             cm_end
	bit   #0x00ff, R13
	jnz            cm_end
	swpb           R13
	bic.b #0x02,   6(R15)
	bit   #0xffff, R13
	jnz            cm_end
	bic.b #0x01,   6(R15)
	jmp            cm_end
cm_err:
	mov.b #0xff,   7(R15)
cm_end:
	mov   R13,     0(R15)
	pop            R6
	pop            R7
	pop            R8
	pop            R9
	pop            R10
	pop            R11
	ret

;the calculator divide function
.global calc_div
.type calc_div, @function
.type cd_slo1, @function
.type cd_slo2, @function
.type cd_slo3, @function
.type cd_slo4, @function
.type cd_ml, @function
.type cd_slaa, @function
.type cd_sbdv1, @function
.type cd_sbdv2, @function
.type cd_sbdv3, @function
.type cd_nr1, @function
.type cd_nr2, @function
.type cd_srrl1, @function
.type cd_srrl2, @function
.type cd_tbg, @function
.type cd_srrb, @function
.type cd_zap, @function
.type cd_err, @function
.type cd_end, @function
calc_div:
	push  R11
	push  R10
	push  R9
	push  R8
	push  R7
	push  R6
	push  R5
	
	mov   2(R15),  R7
	bit   #0xffff, R7
	jz             cd_err
	mov   #0x9999, R6
	sub   R7,      R6
	clrc
	dadd  #0x0001, R6
	
	mov   4(R15),  R8
	clr            R9
	clr            R10
	clr            R11
	clr            R12
	clr            R13
	clr            R5
	
	bit.b #0x20,   6(R15)
	jnz            cd_slo1
	add   #8,      R5     
cd_slo1:
	bit.b #0x08,   6(R15)
	jz             cd_slo2
	add   #8,      R5
cd_slo2:
	bit   #0xffff, R5
	jz             cd_slo4
cd_slo3:
	rla            R8
	rlc            R9
	dec            R5
	jnz            cd_slo3
cd_slo4:
	
	mov   #9,      R14
cd_ml:
	mov   #4,      R5
cd_slaa:
	rla            R8
	rlc            R9
	rlc            R10
	rlc            R11
	rla            R12
	rlc            R13
	jc             cd_err
	dec            R5
	jnz            cd_slaa
cd_sbdv1:
	bit   #0xffff, R11
	jnz            cd_sbdv2
	cmp   R7,      R10
	jlo            cd_sbdv3
cd_sbdv2:
	clrc
	dadd  R6,      R10
	dadd  #0x9999, R11
	clrc
	dadd  #0x0001, R12
	jmp            cd_sbdv1
cd_sbdv3:
	dec            R14
	jnz            cd_ml
	
	bic.b #0x03,   6(R15)
	clr   R14
	mov   #0x0000, R9
	mov   #0xf000, R10
	mov   #0xfff0, R11
	bit.b #0x10,   6(R15)
	jz             cd_nr1
	xor.b #0x01,   6(R15)
	xor   #0xf000, R9
	xor   #0x0f00, R10
	xor   #0x000f, R11
cd_nr1:
	bit.b #0x04,   6(R15)
	jz             cd_nr2
	xor.b #0x01,   6(R15)
	xor   #0xf000, R9
	xor   #0x0f00, R10
	xor   #0x000f, R11
cd_nr2:
	bit   R10,     R13
	jnz            cd_err
	bit   R11,     R13
	jnz            cd_srrl1
	mov   #12,     R5
	bis.b #0x02,   6(R15)
	jmp            cd_srrl2
cd_srrl1:
	mov  #4,       R5
cd_srrl2:
	rla            R12
	rlc            R13
	rlc            R14
	dec            R5
	jnz            cd_srrl2
	and   #0xf000, R12
	cmp   #0x5000, R12
	jlo            cd_zap
	clrc
	dadd  #0x0001, R13
	dadc           R14
	jnz            cd_tbg
	bit   R9,      R13
	jnz            cd_tbg
	jmp            cd_zap
cd_tbg:
	bit.b #0x02,   6(R15)
	jz             cd_err
	bic.b #0x02,   6(R15)
	mov   #8,      R5
cd_srrb:
	rra            R13
	dec            R5
	jnz            cd_srrb
cd_zap:
	bit.b #0x02,   6(R15)
	jz             cd_end
	bit   #0x00ff, R13
	jnz            cd_end
	swpb           R13
	bic.b #0x02,   6(R15)
	bit   #0xffff, R13
	jnz            cd_end
	bic.b #0x01,   6(R15)
	jmp            cd_end
cd_err:
	mov.b #0xff,   7(R15)
cd_end:
	mov   R13,     0(R15)
	pop            R5
	pop            R6
	pop            R7
	pop            R8
	pop            R9
	pop            R10
	pop            R11
	ret
	
	
	