// SilentTimer Action Figure software, v2.1
// definitions.h
// Definitions used by the code.
// 
// 09.06.2020
// Copyright (C) 2014-2017, 2020  Balthasar Szczepański
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

//RTC commands:
#define RTC_EEREAD  0x03
#define RTC_EEWRITE 0x02
#define RTC_EEWRDI  0x04
#define RTC_EEWREN  0x06
#define RTC_SRREAD  0x05
#define RTC_SRWRITE 0x01
#define RTC_READ    0x13
#define RTC_WRITE   0x12
#define RTC_UNLOCK  0x14
#define RTC_IDWRITE 0x32
#define RTC_IDREAD  0x33
#define RTC_CLRRAM  0x54

//SPI function flags:
#define ADDRFLAG  0x01
#define WRITEFLAG 0x02
#define READFLAG  0x04

//7 segment display

#if defined NEW_DISPLAY

#define SEG_A 0xfb
#define SEG_B 0xf7 //    aaaa
#define SEG_C 0xfd //   f    b
#define SEG_D 0x7f //   f    b
#define SEG_E 0xbf //hh  gggg
#define SEG_F 0xfe //   e    c
#define SEG_G 0xdf //   e    c
#define SEG_H 0xef //    dddd 

#define PIN_AB 0x23
#define PIN_CD 0x17
#define PIN_EF 0x60
#define PIN_GH 0x54

#elif defined OLD_HARDWARE

#define SEG_A 0xdf
#define SEG_B 0xef //    aaaa
#define SEG_C 0xbf //   f    b
#define SEG_D 0x7f //   f    b
#define SEG_E 0xfd //hh  gggg
#define SEG_F 0xfb //   e    c
#define SEG_G 0xf7 //   e    c
#define SEG_H 0xfe //    dddd 

#define PIN_AB 0x54
#define PIN_CD 0x67
#define PIN_EF 0x12
#define PIN_GH 0x30

#else

#define SEG_A 0xfd
#define SEG_B 0xef //    aaaa
#define SEG_C 0xfb //   f    b
#define SEG_D 0xfe //   f    b
#define SEG_E 0xdf //hh  gggg
#define SEG_F 0x7f //   e    c
#define SEG_G 0xbf //   e    c
#define SEG_H 0xf7 //    dddd 

#define PIN_AB 0x14
#define PIN_CD 0x20
#define PIN_EF 0x57
#define PIN_GH 0x63

#endif

#define SYMB_0   (SEG_A & SEG_B & SEG_C & SEG_D & SEG_E & SEG_F        )
#define SYMB_1   (        SEG_B & SEG_C                                )
#define SYMB_2   (SEG_A & SEG_B         & SEG_D & SEG_E         & SEG_G)
#define SYMB_3   (SEG_A & SEG_B & SEG_C & SEG_D                 & SEG_G)
#define SYMB_4   (        SEG_B & SEG_C                 & SEG_F & SEG_G)
#define SYMB_5   (SEG_A         & SEG_C & SEG_D         & SEG_F & SEG_G)
#define SYMB_6   (SEG_A         & SEG_C & SEG_D & SEG_E & SEG_F & SEG_G)
#define SYMB_7   (SEG_A & SEG_B & SEG_C                                )
#define SYMB_8   (SEG_A & SEG_B & SEG_C & SEG_D & SEG_E & SEG_F & SEG_G)
#define SYMB_9   (SEG_A & SEG_B & SEG_C                 & SEG_F & SEG_G)
#define SYMB_A   (SEG_A & SEG_B & SEG_C         & SEG_E & SEG_F & SEG_G)
#define SYMB_B   (                SEG_C & SEG_D & SEG_E & SEG_F & SEG_G)
#define SYMB_C   (                        SEG_D & SEG_E         & SEG_G)
#define SYMB_D   (        SEG_B & SEG_C & SEG_D & SEG_E         & SEG_G)
#define SYMB_E   (SEG_A                 & SEG_D & SEG_E & SEG_F & SEG_G)
#define SYMB_F   (SEG_A                         & SEG_E & SEG_F & SEG_G)
#define SYMB_NUL (0xFF)
#define SYMB_o   (SEG_A & SEG_B                         & SEG_F & SEG_G)
#define SYMB_n   (SEG_A & SEG_B                         & SEG_F        )
#define SYMB_g   (SEG_A & SEG_B & SEG_C & SEG_D         & SEG_F & SEG_G)
#define SYMB_U   (                SEG_C & SEG_D & SEG_E                )
#define SYMB_L   (                        SEG_D & SEG_E & SEG_F        )
#define SYMB_M1  (SEG_A & SEG_B                 & SEG_E & SEG_F        )
#define SYMB_M2  (SEG_A & SEG_B & SEG_C                 & SEG_F        )
#define SYMB_I   (                                SEG_E                )  
#define SYMB_MIN (                                                SEG_G)
#define SYMB_DOT (SEG_H)

//operation modes
#define MODE_EDIT 0x80


#define MODE_TIME  0x00
#define MODE_ALARM 0x01
#define MODE_DATE  0x02
#define MODE_UP    0x03
#define MODE_DOWN  0x04
#define MODE_CALIB 0x05

#define MODE_VER   0x06
#define MODE_DEBUG 0x07
#define MODE_RESET 0x08

#define MODE_CALC  0x09


#define MODE_SETTIME  0x00
#define MODE_SETALARM 0x01
#define MODE_SETDATE  0x02
#define MODE_SETYEAR  0x03
#define MODE_SETDOWN  0x04
#define MODE_SETCALIB 0x05

#define MODE_SETDEBUGADDR 0x06
#define MODE_SETDEBUGVAL  0x07

#define MODE_SETCALCNUM  0x08
#define MODE_SETCALCOPER 0x09


//event flags
#define EVENT_RTC 0x01
#define EVENT_A   0x02
#define EVENT_B   0x04
#define EVENT_C   0x08
#define EVENT_RS  0x10

#define EVENTS    0x1f

#define EVENT_ALARM 0x20
#define EVENT_0     0x40

#define ALARMVALUE 0xa5


//rs232
#define RXBUFSIZE 11
#define RXDATASIZE 5

#define TXBUFSIZE 32
#define TXBUFMASK 31

#define BAUDCOUNT 625 //6MHz/625 = 9600Hz (buzz. 4.8kHz)
#define BAUDHALF  (313+200) //determined by oscilloscope observation


unsigned char bcd_add_byte (unsigned char, unsigned char);
void          lpm3         ();
void          calc_add     (unsigned char *);
void          calc_mul     (unsigned char *);
void          calc_div     (unsigned char *);
